import logging

from subprocess import CalledProcessError, PIPE
from urllib2 import HTTPError

from ..context import Context
from ..packages import urlopen, parseChangelog
from ..registration import get_specified_repos

logger = logging.getLogger(__name__)

def __call__ (config):
    context = Context(config, logger)
    
    logger.info("Checking for updates to Cattura Platform")
    
    context.apt_update()
    
    cattura_repos = get_specified_repos(context, config.update_from)    
    if not len(cattura_repos):
        raise Exception("No valid repositories specified")
    
    cattura_packages = {}
    for repo_name, repos in cattura_repos.iteritems():
        cattura_packages[repo_name] = []
        for repo in repos:
            logger.info("Fetching package list from [{}]".format(repo.url))
            repo_packages = repo.packages()
            cattura_packages[repo_name] += repo_packages
            logger.info(" - {} packages found.".format(len(repo_packages)))
    
    logger.info("{} total Cattura packages found.".format(len([package for packages in cattura_packages.itervalues() for package in packages])))
    
    logger.info("Finding packages that need to be updated")
    all_updates = get_all_updates(context)
    
    cattura_updates = {}
    for repo_name, packages in cattura_packages.iteritems():
        updates = filter_updates_for_packages(all_updates, packages)
        
        if not config.nochangelogs:
            logger.info("Collecting changelogs for {}...".format(repo_name))
            updates = collect_changelogs(context, updates, cattura_repos[repo_name])
        
        cattura_updates[repo_name] = updates

    logger.info("{} updates to Cattura packages found.".format(len([update for repo_updates in cattura_updates.itervalues() for update in repo_updates])))
    log_updates(cattura_updates)
    return cattura_updates

def get_all_updates (context):
    """Return all packages that need an update."""
    updates = {}
    for line in context.check_output_chroot(["apt-get", "upgrade", "-s"]).split("\n"):
        if not line.startswith("Inst "):
            continue

        (inst, package, current_version, new_version, etc) = line.split(" ", 4)
        updates[package] = {
            "package": package,
            "current_version": current_version[1:-1],
            "new_version": new_version[1:]
        }
    return updates

def filter_updates_for_packages (updates, packages):
    filtered_updates = []
    for package in packages:
        if package.package in updates:
            update = {}
            update.update(updates[package.package])
            update["package"] = package
            filtered_updates.append(update)
    return filtered_updates

def collect_changelogs (context, updates, repos):
    for update in updates:
        changelog = get_changelog_with_apt(context, update["package"]) or get_changelog_from_repos(update["package"], repos)
        if changelog:
            update["changes"] = parseChangelog(changelog).get_changes_since(update["current_version"])
    return updates

def log_updates (updates_per_repo):
    for repo_name, updates in updates_per_repo.iteritems():
        logger.info("Updates from {}:".format(repo_name))
        for update in updates:
            logger.info(" - {package.package} ({current_version} -> {new_version})".format(**update))
            if "changes" in update:
                for changeset in update["changes"]:
                    logger.info("  - {}".format(changeset.version))
                    for change in changeset.changes:
                        logger.info("   * {}".format(change))

def get_changelog_with_apt (context, package):
    try:
        return context.check_output_chroot(["apt-get", "-qq", "changelog", package.package], stderr=PIPE)
    except CalledProcessError:
        return None

def get_changelog_from_repos (package, repos):
    for repo in repos:
        try:
            return urlopen("{}/{}_{}.changelog".format(repo.url, package.package, package.version)).read()
        except HTTPError: pass
    return None